:- module(lattice_rt, [
		join/3,
		meet/3,
		in/2,
		'.<-.'/2,
		'.->.'/2,
		up_domain/2,
		detach_domain/1,

		aux_project_domain/3,
		aux_entail/4,
		aux_apply_answer/3,
		aux_clean_vars/2
	    ]).


:- use_package(attr).
:- use_module(engine(attributes)).

:- use_module(library(aggregates)).
:- use_module(library(format)).
:- use_module(library(terms_vars)).
:- use_module(library(sets)).

:- include(lattice_syntax).


:- multifile
	domain_meet/3,
	domain_join/3,
	domain_constant/2,
	domain_operation/3,
	in_domain/2,
	
	operation/2.

add_constraint(V1, _, _DomX) :-
	ground(V1), !.
add_constraint(V1, Constraints, NewDom) :-
	(
	    get_attr_local(V1, '$domain'(PrevConst, Dom)) ->
	    insert_constraint(V1, Constraints, PrevConst, FinalCosnt),
	    domain_meet(NewDom, Dom, FinalDom),
	    FinalDom \= bot,
	    put_attr_local(V1, '$domain'(FinalCosnt, FinalDom)),
	    (
		Dom == FinalDom ->
		true
	    ;
		%		print('|'),
		check_constraint(PrevConst)
	    )
	;
	    in_domain(NewDom, _),
	    insert_constraint(V1, Constraints, [], FinalConst),
	    put_attr_local(V1, '$domain'(FinalConst, NewDom))
	).

insert_list_constraint(_,  [],                Rs,        Rs).
insert_list_constraint(V1, [Constraint|Rest], PrevConst, FinalConst) :-
	insert_constraint(V1, Constraint, PrevConst, TempConst),
	insert_list_constraint(V1, Rest, TempConst, FinalConst).
insert_constraint(V1, Constraint, Rs, Rs) :-
	varset(Constraint-V1, Xs),
	Xs == [V1], !.
insert_constraint(_, Constraint, Rs, Final) :-
	insert(Rs, Constraint, Final).
check_constraint([]) :- !.
check_constraint([C|Cs]) :-
	call(C),
	check_constraint(Cs).


%% OPERATORS $$

constraint(Constant, []) :-
	ground(Constant), !.
constraint(V1, Const) :-
	(
	    get_attr_local(V1, '$domain'(Const, _)) ->
	    true
	;
	    Const = []
	).

up_domain(Constant, DomX) :-
	ground(Constant), !,
	(
	    domain_constant(Constant, DomX) ->
	    true
	;
	    fail
	).
up_domain(X, NewDomX) :-
	ground(NewDomX), !,
	(
	    get_attr_local(X, '$domain'(PrevConst, Dom)) ->
	    domain_meet(NewDomX, Dom, FinalDom),
	    FinalDom \= bot,
	    put_attr_local(X, '$domain'(PrevConst, FinalDom)),
	    (
		Dom == FinalDom ->
		true
	    ;
		%		true
		check_constraint(PrevConst)
	    )
	;
	    in_domain(NewDomX, _),
	    put_attr_local(X, '$domain'([], NewDomX))
	).
up_domain(X, DomX) :-
	(
	    get_attr_local(X, '$domain'(_, PrevDom)) ->
	    DomX = PrevDom
	;
	    DomX = top,
	    put_attr_local(X, '$domain'([], DomX))
	).


operation(Op, Var) :-
	list_domain(Var, DomIn),
	domain_operation(Op, DomIn, DomOut),
	list_add_constraint(Op, Var, Var, DomOut).
list_domain([],     []).
list_domain([V|Vs], [D|Ds]) :-
	up_domain(V, D),
	list_domain(Vs, Ds).

list_add_constraint(_,  _,   [],     []).
list_add_constraint(Op, Var, [V|Vs], [D|Ds]) :-
	add_constraint(V, operation(Op, Var), D),
	list_add_constraint(Op, Var, Vs, Ds).



%% lattice constraint - 
V1 in Domain :-
	ground(Domain),
	in_domain(Domain,_), !,
	add_constraint(V1, V1 in Domain, Domain).
	

V1 in V2 :-
	up_domain(V1, DomV1),
	up_domain(V2, DomV2),
	domain_meet(DomV1, DomV2, NewDomV1),
	add_constraint(V1, V1 in V2, NewDomV1),
	add_constraint(V2, V1 in V2, DomV2).

V1 .<-. V2 :-
	ground(V2),
	in_domain(V2, _), !,
	(
	    get_attr_local(V1, '$domain'(Const, _)) ->
	    put_attr_local(V1, '$domain'(Const, V2))
	;
	    put_attr_local(V1, '$domain'([], V2))
	).

V1 .->. V2 :-
	var(V2),
	get_attr_local(V1, '$domain'(_, V2)).


detach_domain(V1) :-
	get_attribute(V1, _), !,
	detach_attribute(V1).
detach_domain(_).

%% join /\ lub
join(A, B, S) :-
	up_domain(A, DomA),
	up_domain(B, DomB),
	domain_join(DomA, DomB, DomS),
	DomS \= bot,
	up_domain(S, DomS).

%% meet \/ glb
meet(A, B, S) :-
	up_domain(A, DomA),
	up_domain(B, DomB),
	domain_meet(DomA, DomB, DomS),
	DomS \= bot,
	up_domain(S, DomS).


%% attr pedricates %%
:- multifile attr_portray_hook/2.
:- multifile attr_unify_hook/2. %% needed
:- multifile attribute_goals/3.

attr_portray_hook(Att, Var) :-
	display(Var), display(', '), display(Att).

attr_unify_hook('$domain'(ConstX, DomX), Y) :-
	get_attr_local(Y, '$domain'(ConstY, DomY)),
	domain_meet(DomX, DomY, FinalDom),
	ord_union(ConstX, ConstY, FinalConst),
	put_attr_local(Y, '$domain'(FinalConst, FinalDom)),
	% check_constraint(FinalConst).
	(
	    DomX == DomY ->
	    true
	;
	    DomX = FinalDom ->
	    check_constraint(ConstY)
	;
	    DomY = FinalDom ->
	    check_constraint(ConstX)
	;
	    check_constraint(FinalConst)
	).

attribute_goals(X, [S|T], T) :-
	get_attr_local(X, '$domain'(Consts, Dom)),
	S = domain([X, Consts, Dom]).
%% attr pedricates %%




%% Aux predicate for tabling interface:

%% DUMP ATTRIBUTES %%
aux_project_domain(Vars, Space, Dom) :-
	dump_store(Vars, Space),
	dump_domain(Vars, Dom).
dump_store(Vars, Space) :-
	project_store(Vars, Space).
dump_domain([],     []).
dump_domain([X|Xs], [D|Ds]) :-
	up_domain(X, D),
	dump_domain(Xs, Ds).


%% ENTAIL %%
aux_entail(SpaceA, DomainA, SpaceB, DomainB) :-
	entail_domain(DomainA, DomainB),
	entail_store(SpaceA, SpaceB).
entail_domain([],     []).
entail_domain([X|Xs], [Y|Ys]) :-
	domain_meet(X, Y, X),
	entail_domain(Xs, Ys).
entail_store([Vars, SpaceA], [Vars, SpaceB]) :-
	sort(SpaceA, SortA),
	sort(SpaceB, SortB),
	ord_subtract(SortA, SortB, []).

%% APPLY ANSWER %%
aux_apply_answer(Vars, Space, Dom) :-
	apply_store(Vars, Space),
	apply_domain(Vars, Dom).
apply_domain([],     []).
apply_domain([X|Xs], [D|Ds]) :-
	check_answer(X, D),
	apply_domain(Xs, Ds).
apply_store(_,    [_,    []]) :- !.
apply_store(Vars, [Vars, Cosntraints]) :-
	call_constraints(Cosntraints).
call_constraints([]).
call_constraints([X|Xs]) :-
	call(X),
	call_constraints(Xs).
check_answer(X, NewDomX) :-
	var(X),
	get_attribute(X, att(X, _, [])), !,
	put_attr_local(X, '$domain'([], NewDomX)).
check_answer(X, NewDomX) :-
	get_attr_local(X, '$domain'(Const, DomX)), !,
	domain_meet(NewDomX, DomX, FinalDomX),
	put_attr_local(X, '$domain'(Const, FinalDomX)).
check_answer(X, NewDomX) :-
	put_attr_local(X, '$domain'([], NewDomX)).


%% PROJECTION SPACE LATTICE (V1 in V2) %%
aux_clean_vars(Vars, [Vars, Space]) :-
	clean_constraints(Vars),
	call_constraints(Space).
clean_constraints([]).
clean_constraints([X|Xs]) :-
	get_attr_local(X, '$domain'(_, DomX)),
	put_attr_local(X, '$domain'([], DomX)),
	clean_constraints(Xs).



project_store(Vars, [Vars, Space]) :-
	sort(Vars, V),
	project(V, Others, Space1),
	simplifier(Space1, Others, Space).

simplifier(Space1, [],     Space1).
simplifier(Space1, [O|Os], FinalSpace) :-
	split_store(in, Space1, O, SpaceLeft, SpaceRight, NewSpace),
	new_constraint(in, SpaceLeft, SpaceRight, NewConstraints),
	ord_union(NewSpace, NewConstraints, NextSpaceA),
	simplifier(NextSpaceA, Os, FinalSpace).

new_constraint(_,  [],     _,  []).
new_constraint(Op, [L|Ls], Rs, NewC) :-
	L =.. [Op, _, V],
	new_constraint_aux(Op, V, Rs, NewC0),
	new_constraint(Op, Ls, Rs, NewC1),
	ord_union(NewC0, NewC1, NewC).

new_constraint_aux(_,  _, [],     []).
new_constraint_aux(Op, V, [R|Rs], [S|Ss]) :-
	R =.. [Op, V2, _],
	S =.. [Op, V2, V],
	V \== V2, !,
	new_constraint_aux(Op, V, Rs, Ss).
new_constraint_aux(Op, V, [_|Rs], Ss) :-
	new_constraint_aux(Op, V, Rs, Ss).


split_store(_,  [],     _O, [],     [], []).
split_store(Op, [S|Ss], O,  [L|Ls], Rs, Ns) :-
	S =.. [Op, O1, V],
	L =.. [Op, O1, V],
	O1 == O, !,
	split_store(Op, Ss, O, Ls, Rs, Ns).
split_store(Op, [S|Ss], O, Ls, [R|Rs], Ns) :-
	S =.. [Op, V, O1],
	R =.. [Op, V, O1],
	O1 == O, !,
	split_store(Op, Ss, O, Ls, Rs, Ns).
split_store(Op, [C|Ss], O, Ls, Rs, [C|Ns]) :-
	split_store(Op, Ss, O, Ls, Rs, Ns).

project(Vars, TotalOthers, Space) :-
	take_in_const(Vars, C0),
	others(Vars, C0, Others),
	project_aux(Vars, C0, Others, [], TotalOthers, Space).

project_aux(_,    C,  [],      O,         O,           C) :- !.
project_aux(Vars, C0, Others0, PreOthers, TotalOthers, Space) :-
	take_in_const(Others0, C1),
	ord_union(C0,   C1,        C2),
	ord_union(Vars, PreOthers, Total0),
	others(Total0, C2, Others1),
	ord_union(PreOthers, Others1, PreOthers1),
	project_aux(Vars, C2, Others1, PreOthers1, TotalOthers, Space).

take_in_const([],     []).
take_in_const([X|Xs], Const) :-
	constraint(X, Cs),
	filter_in(Cs, In_Cs),
	take_in_const(Xs, CXs),
	ord_union(CXs, In_Cs, Const).

filter_in([],     []).
filter_in([X|Xs], [X|Fs]) :-
	\+ \+ X =.. [in, _A, _B], !,
	filter_in(Xs, Fs).
filter_in([_X|Xs], Fs) :-
	filter_in(Xs, Fs).

others(Vars, InConst, Other) :-
	varset(InConst, Total),
	ord_subtract(Total, Vars, Other).
%% PROJECTION SPACE LATTICE (V1 in V2) %%


